from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.helpers import load_yaml_asset


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "9.1"
    _to_version = "9.2"

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document("study_set", [self.move_doi_from_websites])

    def move_doi_from_websites(self, study: dict):
        r"""
        Transforms `websites` values that are dois into curies and moves them into the `associated_dois slot
        Removes the doi website values from the `websites slot.
        
        >>> m = Migrator()
        >>> m.move_doi_from_websites({'id': 123, 'websites': ['a', 'https://doi.org/10.25982/109073.30/1895615'], 'associated_dois': 
        ...     [{'doi_value': 'j', 'doi_provider': 'k', 'doi_category': 'i'}]})
        {'id': 123, 'websites': ['a'], 'associated_dois': [{'doi_value': 'j', 'doi_provider': 'k', 'doi_category': 'i'}, {'doi_value': 'doi:10.25982/109073.30/1895615', 'doi_category': 'dataset_doi', 'doi_provider': 'kbase'}]}
        """

        doi_updates = load_yaml_asset('migrator_from_9_1_to_9_2/websites_dois.yaml')

         # transform websites into doi curies and add to associated_dois slot
        if 'websites' in study:
            websites = study['websites']
            for doi_update in doi_updates:
                if doi_update['doi_website'] in websites:
                    doi_value = 'doi:' + doi_update['doi_website'][doi_update['doi_website'].find('10'):]
                    new_doi = {
                        'doi_value': doi_value,
                        'doi_category': doi_update['doi_cat'],
                        'doi_provider': doi_update['doi_prov']
                    }
                    study.setdefault('associated_dois', []).append(new_doi)

                    # remove doi website from websites slot
                    websites.remove(doi_update['doi_website'])
        
        return study

    


import difflib
import uuid
from typing import List

from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.helpers import load_yaml_asset


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "X"
    _to_version = "PR19_and_PR70"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Read the instrument data into an instance variable up front, for future reference.
        self.instruments: List[dict] = load_yaml_asset("migrator_from_10_2_0_to_11_0_0_part_08/instrument_set.yaml")

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.create_collection(collection_name="instrument_set")
        self.fill_instrument_set()
      
        self.adapter.process_each_document(collection_name="omics_processing_set", pipeline=[self.transform_instrument_slot]) 

    def fill_instrument_set(self):
        r"""
        Fills the instrument_set collection from a YAML file of curated instrument instances
        
        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>>
        >>> database = {"instrument_set": []}  # seeds the database
        >>> adapter = DictionaryAdapter(database=database)
        >>> m = Migrator(adapter=adapter)
        >>> m.fill_instrument_set()
        >>> result = adapter.get_document_having_value_in_field('instrument_set', 'name', value='12T FT-ICR MS')
        >>> result['name']
        '12T FT-ICR MS'
        >>> result['vendor']
        'bruker'
        >>> result['model']
        'solarix_12T'
        >>> result['type']
        'nmdc:Instrument'
        """

        for instrument in self.instruments:
            self.adapter.insert_document("instrument_set", instrument)

    def transform_instrument_slot(self, omics_doc: dict):
        r"""
        Changes the `instrument_name` slot to be `instrument_used`, and replaces the name string value with an identifier for 
        the instrument

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>>
        >>> # Seed the database with instruments representative of those described in `instrument_set.yaml`.
        >>> #
        >>> # Note: These "input" names and the corresponding "output" names below were reviewed by a stakeholder at:
        >>> #       https://github.com/microbiomedata/berkeley-schema-fy24/pull/247#issuecomment-2342009832
        >>> #
        >>> database = dict(instrument_set=[])
        >>> for (id, name) in [
        ...     ('nmdc:inst-14-tjtq1d92', '12T FT-ICR MS'),
        ...     ('nmdc:inst-14-dzs60b60', '15T FT-ICR MS'),
        ...     ('nmdc:inst-14-nstrhv39', '21T FT-ICR MS'),
        ...     ('nmdc:inst-14-mwrrj632', '7T FT-ICR MS'),
        ...     ('nmdc:inst-14-fas8ny90', 'GC-MS'),
        ...     ('nmdc:inst-14-79zxap02', 'Illumina HiSeq'),
        ...     ('nmdc:inst-14-nn4b6k72', 'Illumina HiSeq 2500'),
        ...     ('nmdc:inst-14-xz5tb342', 'Illumina NextSeq 550'),
        ...     ('nmdc:inst-14-xx07be40', 'Illumina NovaSeq'),
        ...     ('nmdc:inst-14-mr4r2w09', 'Illumina NovaSeq 6000'),
        ...     ('nmdc:inst-14-kw06tx33', 'QExactHF03'),
        ...     ('nmdc:inst-14-e32bpm65', 'QExactP02'),
        ...     ('nmdc:inst-14-njkx0e66', 'VOrbiETD04'),
        ... ]:
        ...     instrument_doc = dict(id=id, name=name, model='', vendor='', type='nmdc:Instrument')
        ...     database['instrument_set'].append(instrument_doc)
        >>> adapter = DictionaryAdapter(database=database)
        >>> m = Migrator(adapter=adapter)
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': '12T_FTICR_B'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-tjtq1d92']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': '21T_Agilent'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-nstrhv39']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': '21T Agilent'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-nstrhv39']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': '7T_FT_ICR_MS'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-mwrrj632']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': 'Agilent_GC_MS_01'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-fas8ny90']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': 'Illumina HiSeq'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-79zxap02']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': 'Illumina HiSeq 2500-1TB'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-nn4b6k72']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': 'QExactP02'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-e32bpm65']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': 'QExactHF03'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-kw06tx33']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': 'Illumina NovaSeq S4'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-mr4r2w09']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': 'Illumina NovaSeq'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-xx07be40']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': 'Illumina NextSeq550'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-xz5tb342']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': 'NovaSeq SP'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-mr4r2w09']}
        >>> m.transform_instrument_slot({'id': 'nmdc:omcp-001', 'instrument_name': 'Illumina Illumina HiSeq'})
        {'id': 'nmdc:omcp-001', 'instrument_used': ['nmdc:inst-14-79zxap02']}

        Sanity tests related to `difflib`:
        >>> difflib.get_close_matches("a", ["ab", "cd"], n=1, cutoff=0.25)  # returns list containing the closest match
        ['ab']
        >>> difflib.get_close_matches("x", ["ab", "cd"], n=1, cutoff=0.25)  # returns empty list when no close matches
        []
        """

        if "instrument_name" in omics_doc:
            original_instrument_name = omics_doc["instrument_name"]

            # SPECIAL CASE: If the instrument name contains–but does not end with—"NovaSeq" (e.g. "NovaSeq S4"
            #               or "NovaSeq SP"), map it to "Illumina NovaSeq 6000", specifically.
            if "NovaSeq" in omics_doc['instrument_name'] and not omics_doc['instrument_name'].endswith("NovaSeq"):
                instrument = self.adapter.get_document_having_value_in_field(
                    collection_name="instrument_set", field_name="name", value="Illumina NovaSeq 6000"
                )

                self.logger.info(f"'instrument_name' in '{omics_doc['id']}' is '{omics_doc['instrument_name']}' "
                                 f"and will be mapped to instrument named 'Illumina NovaSeq 6000'")
            else:
                # Get a list of the allowed instrument names.
                #
                # Note: Instead of consulting the database, we consult the YAML file that was used to seed the database.
                #
                instruments = load_yaml_asset("migrator_from_10_2_0_to_11_0_0_part_08/instrument_set.yaml")
                allowed_instrument_names = [instrument["name"] for instrument in instruments]

                # Determine which of the allowed instrument names most closely matches the original one.
                # Reference: https://docs.python.org/3/library/difflib.html#difflib.get_close_matches
                cutoff = 0.25
                matches = difflib.get_close_matches(original_instrument_name, allowed_instrument_names, n=1, cutoff=cutoff)
                if len(matches) > 0:
                    allowed_instrument_name = matches[0]  # this is the most closely-matching, allowed instrument name

                    # Use the allowed instrument whose name most closely matches the original instrument's name.
                    self.logger.info(f"'instrument_name' in '{omics_doc['id']}' is '{omics_doc['instrument_name']}' "
                                     f"and will be mapped to instrument named '{allowed_instrument_name}'")

                    instrument = self.adapter.get_document_having_value_in_field(
                        collection_name="instrument_set", field_name="name", value=allowed_instrument_name
                    )

                else:
                    raise ValueError(f"The 'instrument_name' value ({original_instrument_name}) "
                                     f"on the OmicsProcessing document ({omics_doc['id']}) "
                                     f"does not match any of the allowed instrument names "
                                     f"(with a similarity score of at least {cutoff}).")

            # Map the Omics document to the allowed instrument.
            instrument_id = instrument.get("id")
            omics_doc["instrument_used"] = [instrument_id]

            # Delete the obsolete field from the Omics document.
            del omics_doc["instrument_name"]

        return omics_doc


        


import pprint

import click
import click_log
import logging

import requests
from dotenv import load_dotenv
from pymongo import MongoClient
from pymongo.errors import OperationFailure
from linkml_runtime.dumpers import yaml_dumper
import os

from linkml_runtime.utils.schemaview import SchemaView

logger = logging.getLogger(__name__)
click_log.basic_config(logger)


def load_schema(schema_source):
    """Load schema from a URL or file."""
    return SchemaView(schema_source)


class SchemaHandler:
    def __init__(self, schema_source):
        self.schema_source = schema_source
        self.view = load_schema(schema_source)

    def get_class_slots(self, class_name, include_scalars=False):
        """Retrieve all slots for a given class with an option to include scalar slots."""
        class_slot_objs = self.view.class_induced_slots(class_name)
        class_slot_obj_dict = {s.name: s for s in class_slot_objs}
        class_slot_names = [s.name for s in class_slot_objs]
        acceptable_slots = []
        for i in class_slot_names:
            slot_obj = class_slot_obj_dict[i]
            if slot_obj.multivalued or include_scalars:
                acceptable_slots.append(i)
            else:
                logger.warning(
                    f"Skipping {class_name} slot {i} because it is not multivalued and include_scalars is False.")
        acceptable_slots.sort()
        return acceptable_slots


def determine_exclusive_and_common_elements(set1, set2, set1_name='set 1 only', set2_name='set 2 only',
                                            intersection_name='intersection'):
    """
    Perform arithmetic between two sets (or lists) and return a dictionary detailing elements unique to each set
    and their intersection.

    Parameters:
    set1 (set or list): First set (or list) for comparison.
    set2 (set or list): Second set (or list) for comparison.
    set1_name (str): Label for unique elements in the first set.
    set2_name (str): Label for unique elements in the second set.
    intersection_name (str): Label for the intersection elements.

    Returns:
    dict: A dictionary containing three keys (set1_name, set2_name, intersection_name) mapping to lists of elements.
    """
    # Convert input to sets if they are lists or tuples
    if isinstance(set1, (list, tuple)):
        set1 = set(set1)
    if isinstance(set2, (list, tuple)):
        set2 = set(set2)

    set1_only = set1 - set2
    set2_only = set2 - set1
    intersection = set1.intersection(set2)

    return {
        set1_name: list(set1_only),
        set2_name: list(set2_only),
        intersection_name: list(intersection),
    }


def get_collection_stats(selected_collections, db):
    selected_stats = {}
    for collection_name in sorted(selected_collections):
        try:
            for_selected_stats = {}

            collection_stats = db.command("collstats", collection_name)

            document_count = collection_stats["count"]
            if document_count != 0:
                size_in_bytes = collection_stats["size"]
                avg_bytes_per_doc = size_in_bytes / document_count

                for_selected_stats["document_count"] = document_count
                for_selected_stats["size_in_bytes"] = size_in_bytes
                for_selected_stats["storageSize"] = collection_stats["storageSize"]
                for_selected_stats["avg_bytes_per_doc"] = avg_bytes_per_doc

                selected_stats[collection_name] = for_selected_stats
            else:
                logger.info(f"No documents found in collection '{collection_name}'")
        except OperationFailure as e:
            logger.warning(f"Failed to retrieve statistics for collection '{collection_name}': {e}")

    return selected_stats


def fetch_and_log_schema_info(ctx):
    schema_handler = ctx.obj['SCHEMA_HANDLER']
    database_slots = schema_handler.get_class_slots("Database", include_scalars=True)
    sorted_slots = sorted(database_slots)
    formatted_slots = pprint.pformat(sorted_slots)
    logger.info(f"Database slots:\n{formatted_slots}")

    selected_collections = ctx.obj['SELECTED_COLLECTIONS']

    # Determine the relationship between selected_collections and database_slots
    selected_vs_schema = determine_exclusive_and_common_elements(set1=selected_collections,
                                                                 set1_name="selected_collections only",
                                                                 set2=database_slots, set2_name="Database slots only")

    # Sort the dictionary keys
    selected_vs_schema = {key: selected_vs_schema[key] for key in sorted(selected_vs_schema)}

    logger.info("Relationship between selected collections and schema slots:")
    for key, value in selected_vs_schema.items():
        logger.info(f"{key}: {pprint.pformat(sorted(value))}")

    return database_slots, selected_vs_schema


def get_collection_documents(client_base_url, endpoint_prefix, collection_name, max_docs_per_coll, page_size):
    """
    Fetches all documents in a collection using pagination.

    Parameters:
    client_base_url (str): Base URL of the API.
    endpoint_prefix (str): Path component between the URL and the endpoint name.
    collection_name (str): Name of the collection to fetch documents from.
    max_docs_per_coll (int): Maximum number of documents to retrieve per collection.
    page_size (int): Size of each page in pagination.

    Returns:
    list: List of documents fetched from the collection.
    """
    documents = []
    next_page_token = None

    while len(documents) <= max_docs_per_coll:
        url = f"{client_base_url}/{collection_name}/{endpoint_prefix}?max_page_size={page_size}"

        if next_page_token:
            url += f"&page_token={next_page_token}"
        logger.info(url)

        response = requests.get(url)
        if response.status_code == 200:
            data = response.json()
            resources = data.get("resources", [])
            documents.extend(resources)

            if "next_page_token" in data:
                next_page_token = data["next_page_token"]
            else:
                break
        else:
            logger.error(
                f"Failed to fetch documents from {collection_name}: {response.status_code} - {response.reason}")
            break

    return documents


# Define the main command group and include the global --schema-source option
@click.group()
@click_log.simple_verbosity_option(logger)
@click.option('--schema-source', required=True, help="The source of the schema. Can be a file path or URL.")
@click.option('--selected-collections', multiple=True, type=str,
              help='List of specific collections to fetch data from. If omitted, all collections will be processed.')
@click.option('--output-yaml', type=click.Path(), required=True,
              show_default=True, help="Output file for storing fetched data.")
@click.option('--max-docs-per-coll', default=100, show_default=True,
              help='Maximum number of documents to retrieve per collection')
@click.pass_context
def cli(ctx, schema_source, selected_collections, output_yaml, max_docs_per_coll):
    """Main command group that handles sub-commands for different access methods."""
    ctx.ensure_object(dict)

    ctx.obj['SCHEMA_HANDLER'] = SchemaHandler(schema_source)
    ctx.obj['SELECTED_COLLECTIONS'] = selected_collections
    ctx.obj['OUTPUT_YAML'] = output_yaml
    ctx.obj['MAX_DOCS_PER_COLL'] = max_docs_per_coll


# Sub-command for API access
@cli.command()
@click.option('--client-base-url', default="https://api.microbiomedata.org", show_default=True,
              help='HTTP(S) path to the API server.')
@click.option('--endpoint-prefix', default="nmdcschema", show_default=True,
              help='API path component between the URL and the endpoint name')
@click.option('--page-size', default=200_000, show_default=True,
              help='Size of each page in pagination.')
@click.pass_context
def dump_from_api(ctx, client_base_url, endpoint_prefix, page_size):
    """Sub-command that dumps selected records from the nmdc-runtime API."""
    # click.echo("Warning: This API access method is under development and not fully operational.")
    # logger.warning("This API access method is under development and not fully operational.")

    max_docs_per_coll = ctx.obj['MAX_DOCS_PER_COLL']
    selected_collections = ctx.obj['SELECTED_COLLECTIONS']

    (database_slots, selected_vs_schema) = fetch_and_log_schema_info(ctx)  # just printing for now

    # Make requests call to fetch collection stats
    collection_stats_url = f"{client_base_url}/{endpoint_prefix}/collection_stats"
    response = requests.get(collection_stats_url)

    documents = {}

    if response.status_code == 200:
        collection_stats = response.json()
        logger.info("Collection statistics:")

        # sort collection_stats by the ns value
        collection_stats = sorted(collection_stats, key=lambda x: x.get("ns"))
        for stats in collection_stats:
            ns = stats.get("ns")
            storage_stats = stats.get("storageStats")
            logger.info(f"Namespace: {ns}")
            logger.info(f"Storage Stats:")
            logger.info(pprint.pformat(storage_stats))
            logger.info("------------------------------------")
            as_collection_name = ns.split(".")[-1]
            if as_collection_name in selected_collections:
                documents[as_collection_name] = get_collection_documents(
                    client_base_url,
                    as_collection_name,
                    endpoint_prefix,
                    max_docs_per_coll,
                    page_size,
                )
                logger.info(f"{len(documents) = }")
            else:
                logger.warning(f"Collection '{ns}' is not selected.")
    else:
        logger.error(f"Failed to fetch collection stats: {response.status_code} - {response.reason}")

    output_yaml = ctx.obj['OUTPUT_YAML']

    logger.info(f"Writing data to {output_yaml}")
    yaml_dumper.dump(documents, output_yaml)
    logger.info(f"Data from PyMongo successfully written to {output_yaml}")


@cli.command()
@click.pass_context
@click_log.simple_verbosity_option(logger)
@click.option('--env-file', default='local/.env', show_default=True,
              help='Path to .env file containing MongoDB credentials.', type=click.Path(exists=True))
@click.option('--mongo-db-name', default="nmdc", show_default=True, help='MongoDB database name')
@click.option('--mongo-host', default="localhost", show_default=True,
              help='MongoDB host name/address.')
@click.option('--mongo-port', default=27777, type=int, show_default=True, help='MongoDB port')
@click.option('--admin-db', default=None, help='MongoDB authentication source. Leave blank to not specify.')
@click.option('--auth-mechanism', default=None, show_default=True,
              help='Authentication mechanism for MongoDB connection. Leave blank to not specify.')
@click.option('--direct-connection/--no-direct-connection', default=True,
              help='Whether to use a direct connection to MongoDB. Defaults to direct connection.')
def dump_from_database(ctx, env_file, mongo_db_name, mongo_host, mongo_port,
                       admin_db, auth_mechanism, direct_connection):
    """Sub-command that dumps selected records from NMDC's PyMongo."""
    (database_slots, selected_vs_schema) = fetch_and_log_schema_info(ctx)  # just printing for now

    """Sub-command to exclusively use PyMongo for data fetching."""
    logger.info("Starting data fetch using PyMongo...")

    load_dotenv(env_file, verbose=True)
    mongo_user = os.getenv('SOURCE_MONGO_USER')
    mongo_pw = os.getenv('SOURCE_MONGO_PASS')

    # Construct MongoClient parameters dynamically
    client_params = {
        'host': mongo_host,
        'port': mongo_port,
        'directConnection': direct_connection
    }

    # Optional parameters
    if mongo_user:
        client_params['username'] = mongo_user
    if mongo_pw:
        client_params['password'] = mongo_pw
    if auth_mechanism:
        client_params['authMechanism'] = auth_mechanism
    if admin_db:
        client_params['authSource'] = admin_db

    client = MongoClient(**client_params)

    db = client[mongo_db_name]

    data = {}

    max_docs_per_coll = ctx.obj['MAX_DOCS_PER_COLL']
    selected_collections = ctx.obj['SELECTED_COLLECTIONS']

    # the actionable collections are the intersection of
    #   selected_collections and the collection_names in collection_stats

    # Get collection statistics
    collection_stats = get_collection_stats(database_slots, db)

    logger.info("Statistics for applicable collections:")

    # sort collection_stats by collection name key
    collection_stats = dict(sorted(collection_stats.items()))

    for collection_name, stats in collection_stats.items():
        logger.info(f"{collection_name}: {stats}")
        collection = db[collection_name]

        try:
            documents = list(collection.find(limit=max_docs_per_coll))
            # Remove '_id' field from documents to avoid issues with ObjectId
            for document in documents:
                document.pop('_id', None)  # Remove '_id' if it exists in the document
            if collection_name in selected_collections:
                data[collection_name] = documents
                logger.info(f"Retrieved {len(documents)} documents from collection '{collection_name}'")
            else:
                logger.warning(f"Collection '{collection_name}' was not requested.")

        except OperationFailure as e:
            logger.error(f"Failed to fetch from collection {collection_name}: {e}")

    output_yaml = ctx.obj['OUTPUT_YAML']

    logger.info(f"Writing data to {output_yaml}")
    yaml_dumper.dump(data, output_yaml)
    logger.info(f"Data from PyMongo successfully written to {output_yaml}")


if __name__ == '__main__':
    cli(obj={})  # Ensure the context object is initialized if needed



from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""
    Migrates a database between two schemas.

    Note: This is a "no op" migrator. Its existence serves as documentation that no
          database migration is necessary between the specified schema versions.
    """

    _from_version = "10.5.6"
    _to_version = "10.6.0"

    def upgrade(self) -> None:
        r"""Do nothing."""
        pass



from typing import List
from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    """
    Migrates data from X to PR192, makes the existing extraction_target slot multivalued and change its name to extraction_targets

    Should be run after migrator_from_10_2_0_to_11_0_0_part_09.py.
    """

    def upgrade(self):
        r"""Migrates the database from conforming to the original schema, to conforming to the new schema."""

        self.adapter.process_each_document("extraction_set", [self.rename_extraction_target])

    def rename_extraction_target(self, extraction: dict) -> dict:
        r"""
        Renames the `extraction_target` field to `extraction_targets` and makes it multivalued.

        >>> m = Migrator()
        >>> m.rename_extraction_target({'id': 123})  # no `extraction_target` field
        {'id': 123}
        >>> m.rename_extraction_target({'id': 123, 'extraction_target': 'DNA'})  # test: renames field and casts it as a list value
        {'id': 123, 'extraction_targets': ['DNA']}
        """

        self.logger.info(f"Starting migration of {extraction['id']}")
        if "extraction_target" in extraction:
            extraction["extraction_targets"] = [extraction.pop("extraction_target")]
        return extraction


from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.partials.migrator_from_11_0_3_to_11_1_0 import (
    get_migrator_classes,
)


class Migrator(MigratorBase):
    r"""
    Migrates a database between two schemas.

    Reference: https://pypi.org/project/nmdc-schema/#history
    """

    _from_version = "11.0.3"
    _to_version = "11.1.0" 

    def upgrade(self) -> None:
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        This migrator uses partial migrators. It runs them in the order in which they were designed to be run.
        """

        migrator_classes = get_migrator_classes()
        num_migrators = len(migrator_classes)
        for idx, migrator_class in enumerate(migrator_classes):
            self.logger.info(f"Running migrator {idx + 1} of {num_migrators}")
            self.logger.debug(
                f"Migrating from {migrator_class.get_origin_version()} "
                f"to {migrator_class.get_destination_version()}"
            )
            migrator = migrator_class(adapter=self.adapter, logger=self.logger)
            migrator.upgrade()


from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""Migrates a database between two schemas."""

    _from_version = "11.1.0"
    _to_version = "11.2.0.part_1"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # This is a dictionary of key-value pairs, where each key is the `id` of a `workflow_execution_set` document
        # and the corresponding value is the value we will store in that document's `uses_calibration` field.
        self.calibration_mappings = {}

    def upgrade(self) -> None:
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> database = dict(
        ...     data_generation_set=[
        ...         {'id': 'nmdc:dgms-99-000111', 'type': 'nmdc:MassSpectrometry', 'analyte_category': 'nom', 'has_calibration': 'nmdc:calib-99-000111'},
        ...         {'id': 'nmdc:dgms-99-000222', 'type': 'nmdc:MassSpectrometry', 'analyte_category': 'metabolome', 'has_calibration': 'nmdc:calib-99-000222'},
        ...     ],
        ...     workflow_execution_set=[
        ...         {'id': 'nmdc:wfnom-99-000111', 'type': 'nmdc:NomAnalysis', 'was_informed_by': 'nmdc:dgms-99-000111'},
        ...         {'id': 'nmdc:wfnom-99-000222', 'type': 'nmdc:NomAnalysis', 'was_informed_by': 'nmdc:dgms-99-000222'},
        ...     ],
        ... )
        >>> m = Migrator(adapter=DictionaryAdapter(database=database))
        >>> m.calibration_mappings = {}
        >>> m.upgrade()
        >>> database['data_generation_set'][0]  # test: loses `has_calibration` (and then `generates_calibration`) field
        {'id': 'nmdc:dgms-99-000111', 'type': 'nmdc:MassSpectrometry', 'analyte_category': 'nom'}
        >>> database['data_generation_set'][1]  # test: `has_calibration` field gets renamed to `generates_calibration`
        {'id': 'nmdc:dgms-99-000222', 'type': 'nmdc:MassSpectrometry', 'analyte_category': 'metabolome', 'generates_calibration': 'nmdc:calib-99-000222'}
        >>> database['workflow_execution_set'][0]  # test: gains `uses_calibration` field
        {'id': 'nmdc:wfnom-99-000111', 'type': 'nmdc:NomAnalysis', 'was_informed_by': 'nmdc:dgms-99-000111', 'uses_calibration': 'nmdc:calib-99-000111'}
        >>> database['workflow_execution_set'][1]  # test: gains `uses_calibration` field
        {'id': 'nmdc:wfnom-99-000222', 'type': 'nmdc:NomAnalysis', 'was_informed_by': 'nmdc:dgms-99-000222', 'uses_calibration': 'nmdc:calib-99-000222'}
        """

        self.adapter.process_each_document(
            "data_generation_set",
            [
                self.rename_has_calibration_field,
            ],
        )

        self.adapter.process_each_document(
            "data_generation_set",
            [
                self.determine_calibration_mapping,
            ],
        )

        self.adapter.process_each_document(
            "workflow_execution_set",
            [
                self.apply_calibration_mapping,
            ],
        )

    def rename_has_calibration_field(self, data_generation: dict) -> dict:
        r"""
        Renames the `has_calibration` field to `generates_calibration`, if the document represents
        an instance of the `MassSpectrometry` class.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> m = Migrator(adapter=DictionaryAdapter(database={}))
        >>> m.rename_has_calibration_field({
        ...     'id': 1,
        ...     'type': 'nmdc:MassSpectrometry',
        ... })  # test: lacks `has_calibration`
        {'id': 1, 'type': 'nmdc:MassSpectrometry'}
        >>> m.rename_has_calibration_field({
        ...     'id': 1,
        ...     'type': 'nmdc:MassSpectrometry',
        ...     'has_calibration': 'nmdc:calib-99-abc123',
        ... })  # test: has `has_calibration`
        {'id': 1, 'type': 'nmdc:MassSpectrometry', 'generates_calibration': 'nmdc:calib-99-abc123'}
        >>> m.rename_has_calibration_field({
        ...     'id': 1,
        ...     'type': '__AnythingElse__',
        ...     'has_calibration': 'nmdc:calib-99-abc123',
        ... })  # test: has different `type`
        {'id': 1, 'type': '__AnythingElse__', 'has_calibration': 'nmdc:calib-99-abc123'}
        """
        if data_generation.get("type") == "nmdc:MassSpectrometry":
            if "has_calibration" in data_generation:
                self.logger.info(
                    f"Renaming `has_calibration` field to `generates_calibration` "
                    f"on document having id: {data_generation['id']}"
                )
                data_generation["generates_calibration"] = (
                    data_generation.pop("has_calibration")
                )

        return data_generation

    def determine_calibration_mapping(self, data_generation: dict) -> dict:
        r"""
        If the specified `data_generation_set` document meets certain criteria, adds its `generates_calibration` or
        `has_calibration` value to the instance-level dictionary that will later be used to populate the
        `uses_calibration` fields of `workflow_execution_set` documents; and if the `analyte_category` value of the
        `data_generation_set` document is "nom" (and not "metabolome"), deletes the `generates_calibration` or
        `has_calibration` field from that `data_generation_set` document.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> database = dict(
        ...     workflow_execution_set=[
        ...         {'id': 'nmdc:wfnom-99-000000', 'type': 'nmdc:NomAnalysis', 'was_informed_by': 'nmdc:dgms-99-000111'},
        ...         {'id': 'nmdc:wfnom-99-000111', 'type': 'nmdc:NomAnalysis', 'was_informed_by': 'nmdc:dgms-99-000222'},
        ...     ]
        ... )
        >>> m = Migrator(adapter=DictionaryAdapter(database=database))
        >>> m.calibration_mappings = {}

        # Test: Function deletes `has_calibration` field from input document and adds an item to the mapping dictionary,
        #       preparing us to—eventually—set the associated `workflow_execution_set` document's `uses_calibration`
        #       field to the value that was in the input document's (now deleted) `has_calibration` field.
        >>> m.determine_calibration_mapping({
        ...     'id': 'nmdc:dgms-99-000111',
        ...     'type': 'nmdc:MassSpectrometry',
        ...     'analyte_category': 'nom',
        ...     'has_calibration': 'nmdc:calib-99-000000',
        ... })
        {'id': 'nmdc:dgms-99-000111', 'type': 'nmdc:MassSpectrometry', 'analyte_category': 'nom'}
        >>> len(m.calibration_mappings.items())
        1
        >>> m.calibration_mappings['nmdc:wfnom-99-000000']
        'nmdc:calib-99-000000'

        # Test: Since the `analyte_category` is not `nom`, the input document retains its `has_calibration` field.
        >>> m.determine_calibration_mapping({
        ...     'id': 'nmdc:dgms-99-000222',
        ...     'type': 'nmdc:MassSpectrometry',
        ...     'analyte_category': 'metabolome',
        ...     'has_calibration': 'nmdc:calib-99-000111',
        ... })
        {'id': 'nmdc:dgms-99-000222', 'type': 'nmdc:MassSpectrometry', 'analyte_category': 'metabolome', 'has_calibration': 'nmdc:calib-99-000111'}
        >>> len(m.calibration_mappings.items())
        2
        >>> m.calibration_mappings['nmdc:wfnom-99-000111']
        'nmdc:calib-99-000111'

        # Test: No changes to input document or mappings, since `analyte_category` is neither `nom` nor `metabolome`.
        >>> m.determine_calibration_mapping({
        ...     'id': 'nmdc:dgms-99-000333',
        ...     'type': 'nmdc:MassSpectrometry',
        ...     'analyte_category': 'lipidome',
        ...     'has_calibration': 'nmdc:calib-99-000333',
        ... })
        {'id': 'nmdc:dgms-99-000333', 'type': 'nmdc:MassSpectrometry', 'analyte_category': 'lipidome', 'has_calibration': 'nmdc:calib-99-000333'}
        >>> len(m.calibration_mappings.items())  # no mappings were created
        2

        # Test: No changes to input document or mappings, since input document lacks calibration-related field.
        >>> m.determine_calibration_mapping({
        ...     'id': 'nmdc:dgms-99-000444',
        ...     'type': 'nmdc:MassSpectrometry',
        ...     'analyte_category': 'nom',
        ... })
        {'id': 'nmdc:dgms-99-000444', 'type': 'nmdc:MassSpectrometry', 'analyte_category': 'nom'}
        >>> len(m.calibration_mappings.items())  # no mappings were created
        2

        # Test: No changes to input document or mappings, since no relevant `workflow_execution_set` document exists.
        #       In other words, in our database (at the top of this set of doctests), there is no
        #       `workflow_execution_set` document whose `was_informed_by` field consists of `nmdc:dgms-99-000555`.
        >>> m.determine_calibration_mapping({
        ...     'id': 'nmdc:dgms-99-000555',
        ...     'type': 'nmdc:MassSpectrometry',
        ...     'analyte_category': 'nom',
        ...     'has_calibration': 'nmdc:calib-99-000444',
        ... })
        {'id': 'nmdc:dgms-99-000555', 'type': 'nmdc:MassSpectrometry', 'analyte_category': 'nom', 'has_calibration': 'nmdc:calib-99-000444'}
        >>> len(m.calibration_mappings.items())  # no mappings were created
        2
        """

        # Make handy aliases for some values.
        data_generation_id = data_generation["id"]
        data_generation_analyte_category = data_generation.get("analyte_category")

        # Implement the algorithm discussed with schema maintainers.
        if data_generation_analyte_category == "nom" or data_generation_analyte_category == "metabolome":

            # Determine which calibration-related field this document has, if any.
            calibration_field_name = None
            if "generates_calibration" in data_generation:
                calibration_field_name = "generates_calibration"
            elif "has_calibration" in data_generation:
                calibration_field_name = "has_calibration"
            else:
                self.logger.warning(f"'data_generation_set' document '{data_generation_id}' has neither field")

            if calibration_field_name is not None:

                # Find the `workflow_execution_set` document whose `was_informed_by` field contains this document's `id`
                workflow_execution = self.adapter.get_document_having_value_in_field(
                    collection_name="workflow_execution_set",
                    field_name="was_informed_by",  # confirmed: this is a single valued field, not a multivalued one
                    value=data_generation_id,
                )
                if workflow_execution is not None:

                    # Update the instance-level dictionary, which will be "consumed" by a different method.
                    self.calibration_mappings[workflow_execution["id"]] = data_generation.get(calibration_field_name)

                    # If this `data_generation_set` document's `analyte_category` value was "nom", specifically,
                    # delete the calibration-related field from the document.
                    if data_generation_analyte_category == "nom":
                        data_generation.pop(calibration_field_name)
                else:
                    self.logger.warning(f"No 'workflow_execution_set' document has a "
                                        f"'was_informed_by' value of '{data_generation_id}'")

        return data_generation

    def apply_calibration_mapping(self, workflow_execution: dict) -> dict:
        r"""
        Sets the `uses_calibration` field to the value defined in the instance-level dictionary.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> m = Migrator(adapter=DictionaryAdapter(database={}))
        >>> m.calibration_mappings = {
        ...     'nmdc:wfmgan-99-abc123': 'nmdc:calib-99-def456',
        ...     'nmdc:wfmgan-99-foobar': 'nmdc:calib-99-bazqux',
        ... }
        >>> m.apply_calibration_mapping({'id': 'nmdc:wfmgan-99-abc123', 'type': 'nmdc:MetagenomeAnnotation'})  # id is in mapping dictionary
        {'id': 'nmdc:wfmgan-99-abc123', 'type': 'nmdc:MetagenomeAnnotation', 'uses_calibration': 'nmdc:calib-99-def456'}
        >>> m.apply_calibration_mapping({'id': 'nmdc:wfmgan-00-abc123', 'type': 'nmdc:MetagenomeAnnotation'})  # id is not in mapping dictionary
        {'id': 'nmdc:wfmgan-00-abc123', 'type': 'nmdc:MetagenomeAnnotation'}
        """
        workflow_execution_id = workflow_execution["id"]

        # Check whether this `workflow_execution_set` document's `id` value is a key in our dictionary.
        if workflow_execution_id in self.calibration_mappings.keys():

            # Set this `workflow_execution_set` document's `uses_calibration` field to the value from our dictionary.
            workflow_execution["uses_calibration"] = self.calibration_mappings[workflow_execution_id]

        return workflow_execution
